!MNH_LIC Copyright 1994-2021 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_FAST_RH
INTERFACE
SUBROUTINE ICE4_FAST_RH(LDSOFT, PCOMPUTE, PWETG, &
                       &PRHODREF, PLVFACT, PLSFACT, PPRES, &
                       &PDV, PKA, PCJ, &
                       &PLBDAS, PLBDAG, PLBDAR, PLBDAH, &
                       &PT,  PRVT, PRCT, PRRT, PRIT, PRST, PRGT, PRHT, &
                       &PRCWETH, PRIWETH, PRSWETH, PRGWETH, PRRWETH, &
                       &PRCDRYH, PRIDRYH, PRSDRYH, PRRDRYH, PRGDRYH, PRDRYHG, PRHMLTR, &
                       &PRH_TEND, &
                       &PA_TH, PA_RC, PA_RR, PA_RI, PA_RS, PA_RG, PA_RH)
IMPLICIT NONE
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PWETG    ! 1. where graupel grows in wet mode, 0. elsewhere
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PPRES    ! absolute pressure at t
REAL, DIMENSION(:),       INTENT(IN)    :: PDV      ! Diffusivity of water vapor in the air
REAL, DIMENSION(:),       INTENT(IN)    :: PKA      ! Thermal conductivity of the air
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAG   ! Slope parameter of the graupel   distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAR   ! Slope parameter of the rain      distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAH   ! Slope parameter of the hail      distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PRVT     ! Water vapor m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRIT     ! Pristine ice m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRGT     ! Graupel m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRHT     ! Hail m.r. at t
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCWETH  ! Dry growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRIWETH  ! Dry growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSWETH  ! Dry growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRGWETH  ! Dry growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRWETH  ! Dry growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCDRYH  ! Wet growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRIDRYH  ! Wet growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSDRYH  ! Wet growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRDRYH  ! Wet growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRGDRYH  ! Wet growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRDRYHG  ! Conversion of hailstone into graupel
REAL, DIMENSION(:),       INTENT(INOUT) :: PRHMLTR  ! Melting of the hailstones
REAL, DIMENSION(:, :),    INTENT(INOUT) :: PRH_TEND ! Individual tendencies
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RR
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RI
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RS
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RG
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RH
END SUBROUTINE ICE4_FAST_RH
END INTERFACE
END MODULE MODI_ICE4_FAST_RH
SUBROUTINE ICE4_FAST_RH(LDSOFT, PCOMPUTE, PWETG, &
                       &PRHODREF, PLVFACT, PLSFACT, PPRES, &
                       &PDV, PKA, PCJ, &
                       &PLBDAS, PLBDAG, PLBDAR, PLBDAH, &
                       &PT,  PRVT, PRCT, PRRT, PRIT, PRST, PRGT, PRHT, &
                       &PRCWETH, PRIWETH, PRSWETH, PRGWETH, PRRWETH, &
                       &PRCDRYH, PRIDRYH, PRSDRYH, PRRDRYH, PRGDRYH, PRDRYHG, PRHMLTR, &
                       &PRH_TEND, &
                       &PA_TH, PA_RC, PA_RR, PA_RI, PA_RS, PA_RG, PA_RH)
!!
!!**  PURPOSE
!!    -------
!!      Computes the fast rh process
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!  P. Wautelet 26/04/2019: replace non-standard FLOAT function by REAL function
!  P. Wautelet 29/05/2019: remove PACK/UNPACK intrinsics (to get more performance and better OpenACC support)
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XALPI,XALPW,XBETAI,XBETAW,XGAMW,XCI,XCL,XCPV,XESTT,XGAMI,XLMTT,XLVTT,XMD,XMV,XRV,XTT,XEPSILO
USE MODD_PARAM_ICE,      ONLY: LCONVHG,LEVLIMIT,LNULLWETH,LWETHPOST
USE MODD_RAIN_ICE_DESCR, ONLY: XBG,XBS,XCEXVT,XCXG,XCXH,XCXS,XDH,XRTMIN
USE MODD_RAIN_ICE_PARAM, ONLY: NWETLBDAG,NWETLBDAH,NWETLBDAR,NWETLBDAS,X0DEPH,X1DEPH,XCOLEXGH,XCOLEXIH,XCOLGH,XCOLIH,XCOLEXSH, &
                               XCOLSH,XEX0DEPH,XEX1DEPH,XFGWETH,XFRWETH,XFSWETH,XFWETH,XKER_GWETH,XKER_RWETH,XKER_SWETH,       &
                               XLBGWETH1,XLBGWETH2,XLBGWETH3,XLBRWETH1,XLBRWETH2,XLBRWETH3,XLBSWETH1,XLBSWETH2,XLBSWETH3,      &
                               XWETINTP1G,XWETINTP1H,XWETINTP1R,XWETINTP1S,XWETINTP2G,XWETINTP2H,XWETINTP2R,XWETINTP2S
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
USE MODE_MPPDB
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PWETG    ! 1. where graupel grows in wet mode, 0. elsewhere
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PPRES    ! absolute pressure at t
REAL, DIMENSION(:),       INTENT(IN)    :: PDV      ! Diffusivity of water vapor in the air
REAL, DIMENSION(:),       INTENT(IN)    :: PKA      ! Thermal conductivity of the air
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAG   ! Slope parameter of the graupel   distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAR   ! Slope parameter of the rain      distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAH   ! Slope parameter of the hail      distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PRVT     ! Water vapor m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRIT     ! Pristine ice m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRGT     ! Graupel m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRHT     ! Hail m.r. at t
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCWETH  ! Dry growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRIWETH  ! Dry growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSWETH  ! Dry growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRGWETH  ! Dry growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRWETH  ! Dry growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCDRYH  ! Wet growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRIDRYH  ! Wet growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSDRYH  ! Wet growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRDRYH  ! Wet growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRGDRYH  ! Wet growth of hailstone
REAL, DIMENSION(:),       INTENT(OUT)   :: PRDRYHG  ! Conversion of hailstone into graupel
REAL, DIMENSION(:),       INTENT(INOUT) :: PRHMLTR  ! Melting of the hailstones
REAL, DIMENSION(:, :),    INTENT(INOUT) :: PRH_TEND ! Individual tendencies
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RR
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RI
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RS
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RG
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RH
!
!*       0.2  declaration of local variables
!
INTEGER, PARAMETER :: IRCWETH=1, IRRWETH=2, IRIDRYH=3, IRIWETH=4, IRSDRYH=5, IRSWETH=6, IRGDRYH=7, IRGWETH=8, &
                    & IFREEZ1=9, IFREEZ2=10
!
INTEGER                            :: IGWET
INTEGER                            :: ISIZE
INTEGER                            :: IDX, JJ, JL
INTEGER, DIMENSION(:), allocatable :: I1
INTEGER, DIMENSION(:), allocatable :: IVEC1, IVEC2
LOGICAL, DIMENSION(:), allocatable :: GWET
REAL,    DIMENSION(:), allocatable :: ZHAIL, ZWET, ZMASK, ZWETH, ZDRYH
REAL,    DIMENSION(:), allocatable :: ZVEC1, ZVEC2, ZVEC3
REAL,    DIMENSION(:), allocatable :: ZZW, &
                                      ZRDRYH_INIT, ZRWETH_INIT, &
                                      ZRDRYHG
!
!-------------------------------------------------------------------------------
!
CALL PRINT_MSG(NVERB_ERROR,'GEN','ICE4_FAST_RH','OPENACC: implemented but never tested')

!$acc data present(PCOMPUTE,PWETG,PRHODREF,PLVFACT,PLSFACT,PPRES,PDV,PKA,PCJ, &
!$acc&                PLBDAS,PLBDAG,PLBDAR,PLBDAH,PT,PRVT,PRCT,PRRT,PRIT,PRST,PRGT,PRHT,        &
!$acc&                PRCWETH,PRIWETH,PRSWETH,PRGWETH,PRRWETH,                                  &
!$acc&                PRCDRYH,PRIDRYH,PRSDRYH,PRRDRYH,PRGDRYH,PRDRYHG,PRHMLTR,                  &
!$acc&                PRH_TEND,PA_TH,PA_RC,PA_RR,PA_RI,PA_RS,PA_RG,PA_RH)

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK1D(PCOMPUTE,"ICE4_FAST_RH beg:PCOMPUTE")
  CALL MPPDB_CHECK1D(PWETG,"ICE4_FAST_RH beg:PWETG")
  CALL MPPDB_CHECK1D(PRHODREF,"ICE4_FAST_RH beg:PRHODREF",PRECISION)
  CALL MPPDB_CHECK1D(PLVFACT,"ICE4_FAST_RH beg:PLVFACT",PRECISION)
  CALL MPPDB_CHECK1D(PLSFACT,"ICE4_FAST_RH beg:PLSFACT",PRECISION)
  CALL MPPDB_CHECK1D(PPRES,"ICE4_FAST_RH beg:PPRES",PRECISION)
  CALL MPPDB_CHECK1D(PDV,"ICE4_FAST_RH beg:PDV",PRECISION)
  CALL MPPDB_CHECK1D(PKA,"ICE4_FAST_RH beg:PKA",PRECISION)
  CALL MPPDB_CHECK1D(PCJ,"ICE4_FAST_RH beg:PCJ",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAS,"ICE4_FAST_RH beg:PLBDAS",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAG,"ICE4_FAST_RH beg:PLBDAG",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAR,"ICE4_FAST_RH beg:PLBDAR",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAH,"ICE4_FAST_RH beg:PLBDAH",PRECISION)
  CALL MPPDB_CHECK1D(PT,"ICE4_FAST_RH beg:PT",PRECISION)
  CALL MPPDB_CHECK1D(PRVT,"ICE4_FAST_RH beg:PRVT",PRECISION)
  CALL MPPDB_CHECK1D(PRCT,"ICE4_FAST_RH beg:PRCT",PRECISION)
  CALL MPPDB_CHECK1D(PRRT,"ICE4_FAST_RH beg:PRRT",PRECISION)
  CALL MPPDB_CHECK1D(PRIT,"ICE4_FAST_RH beg:PRIT",PRECISION)
  CALL MPPDB_CHECK1D(PRST,"ICE4_FAST_RH beg:PRST",PRECISION)
  CALL MPPDB_CHECK1D(PRGT,"ICE4_FAST_RH beg:PRGT",PRECISION)
  CALL MPPDB_CHECK1D(PRHT,"ICE4_FAST_RH beg:PRHT",PRECISION)
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRHMLTR,"ICE4_FAST_RH beg:PRHMLTR",PRECISION)
  CALL MPPDB_CHECK2D(PRH_TEND,"ICE4_FAST_RH beg:PRH_TEND",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_FAST_RH beg:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_FAST_RH beg:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RR,"ICE4_FAST_RH beg:PA_RR",PRECISION)
  CALL MPPDB_CHECK1D(PA_RI,"ICE4_FAST_RH beg:PA_RI",PRECISION)
  CALL MPPDB_CHECK1D(PA_RS,"ICE4_FAST_RH beg:PA_RS",PRECISION)
  CALL MPPDB_CHECK1D(PA_RG,"ICE4_FAST_RH beg:PA_RG",PRECISION)
  CALL MPPDB_CHECK1D(PA_RH,"ICE4_FAST_RH beg:PA_RH",PRECISION)
END IF

ISIZE = Size( PRHODREF )

allocate( i1         ( size( prhodref ) ) )
allocate( ivec1      ( size( prhodref ) ) )
allocate( ivec2      ( size( prhodref ) ) )
allocate( gwet       ( size( prhodref ) ) )
allocate( zhail      ( size( prhodref ) ) )
allocate( zwet       ( size( prhodref ) ) )
allocate( zmask      ( size( prhodref ) ) )
allocate( zweth      ( size( prhodref ) ) )
allocate( zdryh      ( size( prhodref ) ) )
allocate( zvec1      ( size( prhodref ) ) )
allocate( zvec2      ( size( prhodref ) ) )
allocate( zvec3      ( size( prhodref ) ) )
allocate( zzw        ( size( prhodref ) ) )
allocate( zrdryh_init( size( prhodref ) ) )
allocate( zrweth_init( size( prhodref ) ) )
allocate( zrdryhg    ( size( prhodref ) ) )

!$acc data create(I1,IVEC1,IVEC2,GWET, ZHAIL, ZWET, ZMASK, ZWETH, ZDRYH, &
!$acc&               ZVEC1,ZVEC2,ZVEC3,ZZW,ZRDRYH_INIT,ZRWETH_INIT,ZRDRYHG)

!
!*       7.2    compute the Wet and Dry growth of hail
!
!$acc kernels
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(7)-PRHT(JL))) * & ! WHERE(PRHT(:)>XRTMIN(7))
           &MAX(0., -SIGN(1., XRTMIN(2)-PRCT(JL))) * & ! WHERE(PRCT(:)>XRTMIN(2))
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRH_TEND(JL, IRCWETH)=ZMASK(JL) * PRH_TEND(JL, IRCWETH)
  ENDDO
ELSE
  PRH_TEND(:, IRCWETH)=0.
  WHERE(ZMASK(:)==1.)
#ifndef MNH_BITREP
    ZZW(:) = PLBDAH(:)**(XCXH-XDH-2.0) * PRHODREF(:)**(-XCEXVT)
#else
    ZZW(:) = BR_POW(PLBDAH(:),XCXH-XDH-2.0) * BR_POW(PRHODREF(:),-XCEXVT)
#endif
    PRH_TEND(:, IRCWETH)=XFWETH * PRCT(:) * ZZW(:)    ! RCWETH
  END WHERE
ENDIF
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(7)-PRHT(JL))) * & ! WHERE(PRHT(:)>XRTMIN(7))
           &MAX(0., -SIGN(1., XRTMIN(4)-PRIT(JL))) * & ! WHERE(PRIT(:)>XRTMIN(4))
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRH_TEND(JL, IRIWETH)=ZMASK(JL) * PRH_TEND(JL, IRIWETH)
    PRH_TEND(JL, IRIDRYH)=ZMASK(JL) * PRH_TEND(JL, IRIDRYH)
  ENDDO
ELSE
  PRH_TEND(:, IRIWETH)=0.
  PRH_TEND(:, IRIDRYH)=0.
  WHERE(ZMASK(:)==1.)
#ifndef MNH_BITREP
    ZZW(:) = PLBDAH(:)**(XCXH-XDH-2.0) * PRHODREF(:)**(-XCEXVT)
    PRH_TEND(:, IRIWETH)=XFWETH * PRIT(:) * ZZW(:)   ! RIWETH
    PRH_TEND(:, IRIDRYH)=PRH_TEND(:, IRIWETH)*(XCOLIH*EXP(XCOLEXIH*(PT(:)-XTT)))   ! RIDRYH
#else
    ZZW(:) = BR_POW(PLBDAH(:),XCXH-XDH-2.0) * BR_POW(PRHODREF(:),-XCEXVT)
    PRH_TEND(:, IRIWETH)=XFWETH * PRIT(:) * ZZW(:)   ! RIWETH
    PRH_TEND(:, IRIDRYH)=PRH_TEND(:, IRIWETH)*(XCOLIH*BR_EXP(XCOLEXIH*(PT(:)-XTT)))   ! RIDRYH
#endif
  END WHERE
ENDIF
!
!*       7.2.1  accretion of aggregates on the hailstones
!
IGWET = 0
!$acc loop private(IDX) independent
DO JJ = 1, SIZE(GWET)
  ZWET(JJ) = MAX(0., -SIGN(1., XRTMIN(7)-PRHT(JJ))) * & ! WHERE(PRHT(:)>XRTMIN(7))
            &MAX(0., -SIGN(1., XRTMIN(5)-PRST(JJ))) * & ! WHERE(PRST(:)>XRTMIN(5))
            &PCOMPUTE(JJ)
  IF (ZWET(JJ)>0) THEN
!$acc atomic capture
    IGWET = IGWET + 1
    IDX = IGWET
!$acc end atomic
    I1(IDX) = JJ
    GWET(JJ) = .TRUE.
  ELSE
    GWET(JJ) = .FALSE.
  END IF
END DO
!$acc end kernels

IF(LDSOFT) THEN
!$acc kernels
  DO JL=1, ISIZE
    PRH_TEND(JL, IRSWETH)=ZWET(JL) * PRH_TEND(JL, IRSWETH)
    PRH_TEND(JL, IRSDRYH)=ZWET(JL) * PRH_TEND(JL, IRSDRYH)
  ENDDO
!$acc end kernels
ELSE
!$acc kernels
  PRH_TEND(:, IRSWETH)=0.
  PRH_TEND(:, IRSDRYH)=0.
!$acc end kernels
  IF(IGWET>0)THEN
!$acc kernels
    !
    !*       7.2.3  select the (PLBDAH,PLBDAS) couplet
    !
    DO JJ = 1, IGWET
      ZVEC1(JJ) = PLBDAH(I1(JJ))
      ZVEC2(JJ) = PLBDAS(I1(JJ))
    END DO
    !
    !*       7.2.4  find the next lower indice for the PLBDAG and for the PLBDAS
    !               in the geometrical set of (Lbda_h,Lbda_s) couplet use to
    !               tabulate the SWETH-kernel
    !
    ZVEC1(1:IGWET) = MAX( 1.00001, MIN( REAL(NWETLBDAH)-0.00001,           &
#ifndef MNH_BITREP
                          XWETINTP1H * LOG( ZVEC1(1:IGWET) ) + XWETINTP2H ) )
#else
                          XWETINTP1H * BR_LOG( ZVEC1(1:IGWET) ) + XWETINTP2H ) )
#endif
    IVEC1(1:IGWET) = INT( ZVEC1(1:IGWET) )
    ZVEC1(1:IGWET) = ZVEC1(1:IGWET) - REAL( IVEC1(1:IGWET) )
    !
    ZVEC2(1:IGWET) = MAX( 1.00001, MIN( REAL(NWETLBDAS)-0.00001,           &
#ifndef MNH_BITREP
                          XWETINTP1S * LOG( ZVEC2(1:IGWET) ) + XWETINTP2S ) )
#else
                          XWETINTP1S * BR_LOG( ZVEC2(1:IGWET) ) + XWETINTP2S ) )
#endif
    IVEC2(1:IGWET) = INT( ZVEC2(1:IGWET) )
    ZVEC2(1:IGWET) = ZVEC2(1:IGWET) - REAL( IVEC2(1:IGWET) )
    !
    !*       7.2.5  perform the bilinear interpolation of the normalized
    !               SWETH-kernel
    !
    DO JJ = 1,IGWET
      ZVEC3(JJ) = (  XKER_SWETH(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                   - XKER_SWETH(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                                 * ZVEC1(JJ) &
                  - ( XKER_SWETH(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                  - XKER_SWETH(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         * (ZVEC1(JJ) - 1.0)
    END DO
    ZZW(:) = 0.
    DO JJ = 1, IGWET
      ZZW(I1(JJ)) = ZVEC3(JJ)
    END DO
    !
    WHERE(GWET(:))
#ifndef MNH_BITREP
      PRH_TEND(:, IRSWETH)=XFSWETH*ZZW(:)                       & ! RSWETH
                    *( PLBDAS(:)**(XCXS-XBS) )*( PLBDAH(:)**XCXH )  &
                       *( PRHODREF(:)**(-XCEXVT-1.) )               &
                       *( XLBSWETH1/( PLBDAH(:)**2              ) + &
                          XLBSWETH2/( PLBDAH(:)   * PLBDAS(:)   ) + &
                          XLBSWETH3/(               PLBDAS(:)**2) )
      PRH_TEND(:, IRSDRYH)=PRH_TEND(:, IRSWETH)*(XCOLSH*EXP(XCOLEXSH*(PT(:)-XTT)))
#else
      PRH_TEND(:, IRSWETH)=XFSWETH*ZZW(:)                       & ! RSWETH
                    * BR_POW(PLBDAS(:),XCXS-XBS) * BR_POW(PLBDAH(:),XCXH)  &
                       * BR_POW(PRHODREF(:),-XCEXVT-1.)               &
                       *( XLBSWETH1/ BR_P2(PLBDAH(:))               + &
                          XLBSWETH2/( PLBDAH(:)   * PLBDAS(:)   )   + &
                          XLBSWETH3/               BR_P2(PLBDAS(:)) )
      PRH_TEND(:, IRSDRYH)=PRH_TEND(:, IRSWETH)*(XCOLSH*BR_EXP(XCOLEXSH*(PT(:)-XTT)))
#endif
    END WHERE
!$acc end kernels
  ENDIF
ENDIF
!
!*       7.2.6  accretion of graupeln on the hailstones
!
!$acc kernels
IGWET = 0
!$acc loop private(IDX) independent
DO JJ = 1, SIZE(GWET)
  ZWET(JJ)=MAX(0., -SIGN(1., XRTMIN(7)-PRHT(JJ))) * & ! WHERE(PRHT(:)>XRTMIN(7))
          &MAX(0., -SIGN(1., XRTMIN(6)-PRGT(JJ))) * & ! WHERE(PRGT(:)>XRTMIN(6))
          &PCOMPUTE(JJ)
  IF (ZWET(JJ)>0) THEN
!$acc atomic capture
    IGWET = IGWET + 1
    IDX = IGWET
!$acc end atomic
    I1(IDX) = JJ
    GWET(JJ) = .TRUE.
  ELSE
    GWET(JJ) = .FALSE.
  END IF
END DO
!$acc end kernels

IF(LDSOFT) THEN
!$acc kernels
  DO JL=1, ISIZE
    PRH_TEND(JL, IRGWETH)=ZWET(JL) * PRH_TEND(JL, IRGWETH)
    PRH_TEND(JL, IRGDRYH)=ZWET(JL) * PRH_TEND(JL, IRGDRYH)
  ENDDO
!$acc end kernels
ELSE
!$acc kernels
  PRH_TEND(:, IRGWETH)=0.
  PRH_TEND(:, IRGDRYH)=0.
!$acc end kernels
  IF(IGWET>0)THEN
!$acc kernels
    !
    !*       7.2.8  select the (PLBDAH,PLBDAG) couplet
    !
    DO JJ = 1, IGWET
      ZVEC1(JJ) = PLBDAH(I1(JJ))
      ZVEC2(JJ) = PLBDAG(I1(JJ))
    END DO
    !
    !*       7.2.9  find the next lower indice for the PLBDAH and for the PLBDAG
    !               in the geometrical set of (Lbda_h,Lbda_g) couplet use to
    !               tabulate the GWETH-kernel
    !
    ZVEC1(1:IGWET) = MAX( 1.00001, MIN( REAL(NWETLBDAG)-0.00001,           &
#ifndef MNH_BITREP
                          XWETINTP1H * LOG( ZVEC1(1:IGWET) ) + XWETINTP2H ) )
#else
                          XWETINTP1H * BR_LOG( ZVEC1(1:IGWET) ) + XWETINTP2H ) )
#endif
    IVEC1(1:IGWET) = INT( ZVEC1(1:IGWET) )
    ZVEC1(1:IGWET) = ZVEC1(1:IGWET) - REAL( IVEC1(1:IGWET) )
    !
    ZVEC2(1:IGWET) = MAX( 1.00001, MIN( REAL(NWETLBDAG)-0.00001,           &
#ifndef MNH_BITREP
                          XWETINTP1G * LOG( ZVEC2(1:IGWET) ) + XWETINTP2G ) )
#else
                          XWETINTP1G * BR_LOG( ZVEC2(1:IGWET) ) + XWETINTP2G ) )
#endif
    IVEC2(1:IGWET) = INT( ZVEC2(1:IGWET) )
    ZVEC2(1:IGWET) = ZVEC2(1:IGWET) - REAL( IVEC2(1:IGWET) )
    !
    !*       7.2.10 perform the bilinear interpolation of the normalized
    !               GWETH-kernel
    !
    DO JJ = 1,IGWET
      ZVEC3(JJ) = (  XKER_GWETH(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                   - XKER_GWETH(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                                 * ZVEC1(JJ) &
                - (  XKER_GWETH(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                   - XKER_GWETH(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                        * (ZVEC1(JJ) - 1.0)
    END DO
    ZZW(:) = 0.
    DO JJ = 1, IGWET
      ZZW(I1(JJ)) = ZVEC3(JJ)
    END DO
    !
    WHERE(GWET(:))
#ifndef MNH_BITREP
      PRH_TEND(:, IRGWETH)=XFGWETH*ZZW(:)                       & ! RGWETH
                    *( PLBDAG(:)**(XCXG-XBG) )*( PLBDAH(:)**XCXH )  &
                       *( PRHODREF(:)**(-XCEXVT-1.) )               &
                       *( XLBGWETH1/( PLBDAH(:)**2              ) + &
                          XLBGWETH2/( PLBDAH(:)   * PLBDAG(:)   ) + &
                          XLBGWETH3/(               PLBDAG(:)**2) )
#else
      PRH_TEND(:, IRGWETH)=XFGWETH*ZZW(:)                                & ! RGWETH
                    * BR_POW(PLBDAG(:),XCXG-XBG)*BR_POW(PLBDAH(:),XCXH)  &
                       * BR_POW(PRHODREF(:),-XCEXVT-1.)                  &
                       *( XLBGWETH1/ BR_P2(PLBDAH(:))                  + &
                          XLBGWETH2/( PLBDAH(:)   * PLBDAG(:)   )      + &
                          XLBGWETH3/               BR_P2(PLBDAG(:)) )
#endif
      PRH_TEND(:, IRGDRYH)=PRH_TEND(:, IRGWETH)
    END WHERE
    !When graupel grows in wet mode, graupel is wet (!) and collection efficiency must remain the same
    WHERE(GWET(:) .AND. .NOT. PWETG(:)==1.)
#ifndef MNH_BITREP
      PRH_TEND(:, IRGDRYH)=PRH_TEND(:, IRGDRYH)*(XCOLGH*EXP(XCOLEXGH*(PT(:)-XTT)))
#else
      PRH_TEND(:, IRGDRYH)=PRH_TEND(:, IRGDRYH)*(XCOLGH*BR_EXP(XCOLEXGH*(PT(:)-XTT)))
#endif
    END WHERE
!$acc end kernels
  END IF
ENDIF
!$acc kernels
!
!*       7.2.11  accretion of raindrops on the hailstones
!
IGWET = 0
!$acc loop private(IDX) independent
DO JJ = 1, SIZE(GWET)
  ZWET(JJ)=MAX(0., -SIGN(1., XRTMIN(7)-PRHT(JJ))) * & ! WHERE(PRHT(:)>XRTMIN(7))
          &MAX(0., -SIGN(1., XRTMIN(3)-PRRT(JJ))) * & ! WHERE(PRRT(:)>XRTMIN(3))
          &PCOMPUTE(JJ)
  IF (ZWET(JJ)>0) THEN
!$acc atomic capture
    IGWET = IGWET + 1
    IDX = IGWET
!$acc end atomic
    I1(IDX) = JJ
    GWET(JJ) = .TRUE.
  ELSE
    GWET(JJ) = .FALSE.
  END IF
END DO
!$acc end kernels

IF(LDSOFT) THEN
!$acc kernels
  DO JL=1, ISIZE
    PRH_TEND(JL, IRRWETH)=ZWET(JL) * PRH_TEND(JL, IRRWETH)
  ENDDO
!$acc end kernels
ELSE
!$acc kernels
  PRH_TEND(:, IRRWETH)=0.
!$acc end kernels
  IF(IGWET>0)THEN
!$acc kernels
    !
    !*       7.2.12  select the (PLBDAH,PLBDAR) couplet
    !
    DO JJ = 1, IGWET
      ZVEC1(JJ) = PLBDAH(I1(JJ))
      ZVEC2(JJ) = PLBDAR(I1(JJ))
    END DO
    !
    !*       7.2.13 find the next lower indice for the PLBDAH and for the PLBDAR
    !               in the geometrical set of (Lbda_h,Lbda_r) couplet use to
    !               tabulate the RWETH-kernel
    !
    ZVEC1(1:IGWET)=MAX(1.00001, MIN( REAL(NWETLBDAH)-0.00001,           &
#ifndef MNH_BITREP
                          XWETINTP1H*LOG(ZVEC1(1:IGWET))+XWETINTP2H))
#else
                          XWETINTP1H*BR_LOG(ZVEC1(1:IGWET))+XWETINTP2H))
#endif
    IVEC1(1:IGWET)=INT(ZVEC1(1:IGWET))
    ZVEC1(1:IGWET)=ZVEC1(1:IGWET)-REAL(IVEC1(1:IGWET))
    !
    ZVEC2(1:IGWET)=MAX(1.00001, MIN( REAL(NWETLBDAR)-0.00001,           &
#ifndef MNH_BITREP
                          XWETINTP1R*LOG(ZVEC2(1:IGWET))+XWETINTP2R))
#else
                          XWETINTP1R*BR_LOG(ZVEC2(1:IGWET))+XWETINTP2R))
#endif
    IVEC2(1:IGWET)=INT(ZVEC2(1:IGWET))
    ZVEC2(1:IGWET)=ZVEC2(1:IGWET)-REAL(IVEC2(1:IGWET))
    !
    !*       7.2.14 perform the bilinear interpolation of the normalized
    !               RWETH-kernel
    !
    DO JJ=1, IGWET
      ZVEC3(JJ)= (  XKER_RWETH(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RWETH(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                                  * ZVEC1(JJ) &
                 - (  XKER_RWETH(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RWETH(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         *(ZVEC1(JJ) - 1.0)
    END DO
    ZZW(:) = 0.
    DO JJ = 1, IGWET
      ZZW(I1(JJ)) = ZVEC3(JJ)
    END DO
    !
    WHERE(GWET(:))
#ifndef MNH_BITREP
      PRH_TEND(:, IRRWETH) = XFRWETH*ZZW(:)                    & ! RRWETH
                        *( PLBDAR(:)**(-4) )*( PLBDAH(:)**XCXH ) &
                               *( PRHODREF(:)**(-XCEXVT-1.) )   &
                    *( XLBRWETH1/( PLBDAH(:)**2              ) + &
                       XLBRWETH2/( PLBDAH(:)   * PLBDAR(:)   ) + &
                       XLBRWETH3/(               PLBDAR(:)**2) )
#else
      PRH_TEND(:, IRRWETH) = XFRWETH*ZZW(:)                        & ! RRWETH
                        / BR_P4(PLBDAR(:))* BR_POW(PLBDAH(:),XCXH) &
                               * BR_POW(PRHODREF(:),-XCEXVT-1.)    &
                    *( XLBRWETH1/ BR_P2(PLBDAH(:))               + &
                       XLBRWETH2/( PLBDAH(:)   * PLBDAR(:)   )   + &
                       XLBRWETH3/               BR_P2(PLBDAR(:)) )
#endif
    END WHERE
!$acc end kernels
  ENDIF
ENDIF
!$acc kernels
!
DO JL=1, ISIZE
  ZRDRYH_INIT(JL)=PRH_TEND(JL, IRCWETH)+PRH_TEND(JL, IRIDRYH)+ &
                 &PRH_TEND(JL, IRSDRYH)+PRH_TEND(JL, IRRWETH)+PRH_TEND(JL, IRGDRYH)
ENDDO
!
!*       7.3    compute the Wet growth of hail
!
DO JL=1, ISIZE
  ZHAIL(JL)=MAX(0., -SIGN(1., XRTMIN(7)-PRHT(JL))) * & ! WHERE(PRHT(:)>XRTMIN(7))
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRH_TEND(JL, IFREEZ1)=ZHAIL(JL) * PRH_TEND(JL, IFREEZ1)
    PRH_TEND(JL, IFREEZ2)=ZHAIL(JL) * PRH_TEND(JL, IFREEZ2)
  ENDDO
ELSE
  DO JL=1, ISIZE
    PRH_TEND(JL, IFREEZ1)=PRVT(JL)*PPRES(JL)/(XEPSILO+PRVT(JL)) ! Vapor pressure
  ENDDO
  IF(LEVLIMIT) THEN
    WHERE(ZHAIL(:)==1.)
#ifndef MNH_BITREP
      PRH_TEND(:, IFREEZ1)=MIN(PRH_TEND(:, IFREEZ1), EXP(XALPI-XBETAI/PT(:)-XGAMI*ALOG(PT(:)))) ! min(ev, es_i(T))
#else
      PRH_TEND(:, IFREEZ1)=MIN(PRH_TEND(:, IFREEZ1), BR_EXP(XALPI-XBETAI/PT(:)-XGAMI*BR_LOG(PT(:)))) ! min(ev, es_i(T))
#endif
    END WHERE
  ENDIF
  PRH_TEND(:, IFREEZ2)=0.
  WHERE(ZHAIL(:)==1.)
    PRH_TEND(:, IFREEZ1)=PKA(:)*(XTT-PT(:)) +                              &
             (PDV(:)*(XLVTT+(XCPV-XCL)*(PT(:)-XTT)) &
                           *(XESTT-PRH_TEND(:, IFREEZ1))/(XRV*PT(:))           )
#ifndef MNH_BITREP
    PRH_TEND(:, IFREEZ1)=PRH_TEND(:, IFREEZ1)* ( X0DEPH*       PLBDAH(:)**XEX0DEPH +     &
                           X1DEPH*PCJ(:)*PLBDAH(:)**XEX1DEPH )/ &
#else
    PRH_TEND(:, IFREEZ1)=PRH_TEND(:, IFREEZ1)* ( X0DEPH*       BR_POW(PLBDAH(:),XEX0DEPH) +     &
                           X1DEPH*PCJ(:)*BR_POW(PLBDAH(:),XEX1DEPH) )/ &
#endif
                          ( PRHODREF(:)*(XLMTT-XCL*(XTT-PT(:))) )
    PRH_TEND(:, IFREEZ2)=(PRHODREF(:)*(XLMTT+(XCI-XCL)*(XTT-PT(:)))   ) / &
                          ( PRHODREF(:)*(XLMTT-XCL*(XTT-PT(:))) )
  END WHERE
ENDIF
DO JL=1, ISIZE
  !We must agregate, at least, the cold species
  ZRWETH_INIT(JL)=ZHAIL(JL) * MAX(PRH_TEND(JL, IRIWETH)+PRH_TEND(JL, IRSWETH)+PRH_TEND(JL, IRGWETH), &
                                 &MAX(0., PRH_TEND(JL, IFREEZ1) + &
                                         &PRH_TEND(JL, IFREEZ2) * ( &
                     &PRH_TEND(JL, IRIWETH)+PRH_TEND(JL, IRSWETH)+PRH_TEND(JL, IRGWETH) )))
ENDDO
!
!*       7.4    Select Wet or Dry case
!
!Wet case
DO JL=1, ISIZE
  ZWETH(JL) = ZHAIL(JL) * &
            & MAX(0., SIGN(1., MAX(0., ZRDRYH_INIT(JL)-PRH_TEND(JL, IRIDRYH)-PRH_TEND(JL, IRSDRYH)-PRH_TEND(JL, IRGDRYH)) - &
                              &MAX(0., ZRWETH_INIT(JL)-PRH_TEND(JL, IRIWETH)-PRH_TEND(JL, IRSWETH)-PRH_TEND(JL, IRGWETH))))
ENDDO
IF(LNULLWETH) THEN
  DO JL=1, ISIZE
    ZWETH(JL) = ZWETH(JL) * MAX(0., -SIGN(1., -ZRDRYH_INIT(JL))) ! WHERE(ZRDRYH_INIT(:)>0.)
  ENDDO
ELSE
  DO JL=1, ISIZE
    ZWETH(JL) = ZWETH(JL) * MAX(0., -SIGN(1., -ZRWETH_INIT(JL))) ! WHERE(ZRWETH_INIT(:)>0.)
  ENDDO
ENDIF
IF(.NOT. LWETHPOST) THEN
  DO JL=1, ISIZE
    ZWETH(JL) = ZWETH(JL) * MAX(0., -SIGN(1., PT(JL)-XTT)) ! WHERE(PT(:)<XTT)
  ENDDO
ENDIF
DO JL=1, ISIZE
  ZDRYH(JL) = ZHAIL(JL) * &
            & MAX(0., -SIGN(1., PT(JL)-XTT)) * & ! WHERE(PT(:)<XTT)
            & MAX(0., -SIGN(1., 1.E-20-ZRDRYH_INIT(JL))) * & !WHERE(ZRDRYH_INIT(:)>0.)
            & MAX(0., -SIGN(1., MAX(0., ZRDRYH_INIT(JL)-PRH_TEND(JL, IRIDRYH)-PRH_TEND(JL, IRSDRYH)) - &
                               &MAX(0., ZRWETH_INIT(JL)-PRH_TEND(JL, IRIWETH)-PRH_TEND(JL, IRSWETH))))
ENDDO
!
ZRDRYHG(:)=0.
IF(LCONVHG)THEN
  WHERE(ZDRYH(:)==1.)
    ZRDRYHG(:)=ZRDRYH_INIT(:)*ZRWETH_INIT(:)/(ZRDRYH_INIT(:)+ZRWETH_INIT(:))
  END WHERE
ENDIF
DO JL=1, ISIZE
  PRCWETH(JL) = ZWETH(JL) * PRH_TEND(JL, IRCWETH)
  PRIWETH(JL) = ZWETH(JL) * PRH_TEND(JL, IRIWETH)
  PRSWETH(JL) = ZWETH(JL) * PRH_TEND(JL, IRSWETH)
  PRGWETH(JL) = ZWETH(JL) * PRH_TEND(JL, IRGWETH)
  !Collected minus aggregated
  PRRWETH(JL) = ZWETH(JL) * (ZRWETH_INIT(JL) - PRH_TEND(JL, IRIWETH) - &
                             PRH_TEND(JL, IRSWETH) - PRH_TEND(JL, IRGWETH) - &
                             PRH_TEND(JL, IRCWETH))

  PRCDRYH(JL) = ZDRYH(JL) * PRH_TEND(JL, IRCWETH)
  PRIDRYH(JL) = ZDRYH(JL) * PRH_TEND(JL, IRIDRYH)
  PRSDRYH(JL) = ZDRYH(JL) * PRH_TEND(JL, IRSDRYH)
  PRRDRYH(JL) = ZDRYH(JL) * PRH_TEND(JL, IRRWETH)
  PRGDRYH(JL) = ZDRYH(JL) * PRH_TEND(JL, IRGDRYH)
  PRDRYHG(JL) = ZDRYH(JL) * ZRDRYHG(JL)

  PA_RC(JL) = PA_RC(JL) - PRCWETH(JL)
  PA_RI(JL) = PA_RI(JL) - PRIWETH(JL)
  PA_RS(JL) = PA_RS(JL) - PRSWETH(JL)
  PA_RG(JL) = PA_RG(JL) - PRGWETH(JL)
  PA_RH(JL) = PA_RH(JL) + PRCWETH(JL)+PRIWETH(JL)+PRSWETH(JL)+PRGWETH(JL)+PRRWETH(JL)
  PA_RR(JL) = PA_RR(JL) - PRRWETH(JL)
  PA_TH(JL) = PA_TH(JL) + (PRRWETH(JL)+PRCWETH(JL))*(PLSFACT(JL)-PLVFACT(JL))
  PA_RC(JL) = PA_RC(JL) - PRCDRYH(JL)
  PA_RI(JL) = PA_RI(JL) - PRIDRYH(JL)
  PA_RS(JL) = PA_RS(JL) - PRSDRYH(JL)
  PA_RR(JL) = PA_RR(JL) - PRRDRYH(JL)
  PA_RG(JL) = PA_RG(JL) - PRGDRYH(JL) + PRDRYHG(JL)
  PA_RH(JL) = PA_RH(JL) + PRCDRYH(JL)+PRIDRYH(JL)+PRSDRYH(JL)+&
                         &PRRDRYH(JL)+PRGDRYH(JL) - PRDRYHG(JL)
  PA_TH(JL) = PA_TH(JL) + (PRCDRYH(JL)+PRRDRYH(JL))*(PLSFACT(JL)-PLVFACT(JL))
ENDDO
!
!*       7.5    Melting of the hailstones
!
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(7)-PRHT(JL))) * & ! WHERE(PRHT(:)>XRTMIN(7))
           &MAX(0., -SIGN(1., XTT-PT(JL))) * & ! WHERE(PT(:)>XTT)
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRHMLTR(JL)=ZMASK(JL)*PRHMLTR(JL)
  ENDDO
ELSE
  DO JL=1, ISIZE
    PRHMLTR(JL) = ZMASK(JL)* PRVT(JL)*PPRES(JL)/(XEPSILO+PRVT(JL)) ! Vapor pressure
  ENDDO
  IF(LEVLIMIT) THEN
    WHERE(ZMASK(:)==1.)
#ifndef MNH_BITREP
      PRHMLTR(:)=MIN(PRHMLTR(:), EXP(XALPW-XBETAW/PT(:)-XGAMW*ALOG(PT(:)))) ! min(ev, es_w(T))
#else
      PRHMLTR(:)=MIN(PRHMLTR(:), BR_EXP(XALPW-XBETAW/PT(:)-XGAMW*BR_LOG(PT(:)))) ! min(ev, es_w(T))
#endif
    END WHERE
  ENDIF
  DO JL=1, ISIZE
    PRHMLTR(JL) = ZMASK(JL)* (PKA(JL)*(XTT-PT(JL)) +                              &
           ( PDV(JL)*(XLVTT + ( XCPV - XCL ) * ( PT(JL) - XTT )) &
                           *(XESTT-PRHMLTR(JL))/(XRV*PT(JL))         ))
  ENDDO
  WHERE(ZMASK(:)==1.)
    !
    ! compute RHMLTR
    !
    PRHMLTR(:)  = MAX( 0.0,( -PRHMLTR(:) *                     &
#ifndef MNH_BITREP
                           ( X0DEPH*       PLBDAH(:)**XEX0DEPH +     &
                             X1DEPH*PCJ(:)*PLBDAH(:)**XEX1DEPH ) -   &
#else
                           ( X0DEPH*       BR_POW(PLBDAH(:),XEX0DEPH) +     &
                             X1DEPH*PCJ(:)*BR_POW(PLBDAH(:),XEX1DEPH) ) -   &
#endif
                         ( PRH_TEND(:, IRCWETH)+PRH_TEND(:, IRRWETH) )*        &
                               ( PRHODREF(:)*XCL*(XTT-PT(:))) ) /    &
                                             ( PRHODREF(:)*XLMTT ) )
  END WHERE
END IF
DO JL=1, ISIZE
  PA_RR(JL) = PA_RR(JL) + PRHMLTR(JL)
  PA_RH(JL) = PA_RH(JL) - PRHMLTR(JL)
  PA_TH(JL) = PA_TH(JL) - PRHMLTR(JL)*(PLSFACT(JL)-PLVFACT(JL))
ENDDO
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRHMLTR,"ICE4_FAST_RH end:PRHMLTR",PRECISION)
  CALL MPPDB_CHECK2D(PRH_TEND,"ICE4_FAST_RH end:PRH_TEND",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_FAST_RH end:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_FAST_RH end:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RR,"ICE4_FAST_RH end:PA_RR",PRECISION)
  CALL MPPDB_CHECK1D(PA_RI,"ICE4_FAST_RH end:PA_RI",PRECISION)
  CALL MPPDB_CHECK1D(PA_RS,"ICE4_FAST_RH end:PA_RS",PRECISION)
  CALL MPPDB_CHECK1D(PA_RG,"ICE4_FAST_RH end:PA_RG",PRECISION)
  CALL MPPDB_CHECK1D(PA_RH,"ICE4_FAST_RH end:PA_RH",PRECISION)
  !Check all OUT arrays
  CALL MPPDB_CHECK1D(PRCWETH,"ICE4_FAST_RH end:PRCWETH",PRECISION)
  CALL MPPDB_CHECK1D(PRIWETH,"ICE4_FAST_RH end:PRIWETH",PRECISION)
  CALL MPPDB_CHECK1D(PRSWETH,"ICE4_FAST_RH end:PRSWETH",PRECISION)
  CALL MPPDB_CHECK1D(PRGWETH,"ICE4_FAST_RH end:PRGWETH",PRECISION)
  CALL MPPDB_CHECK1D(PRRWETH,"ICE4_FAST_RH end:PRRWETH",PRECISION)
  CALL MPPDB_CHECK1D(PRCDRYH,"ICE4_FAST_RH end:PRCDRYH",PRECISION)
  CALL MPPDB_CHECK1D(PRIDRYH,"ICE4_FAST_RH end:PRIDRYH",PRECISION)
  CALL MPPDB_CHECK1D(PRSDRYH,"ICE4_FAST_RH end:PRSDRYH",PRECISION)
  CALL MPPDB_CHECK1D(PRRDRYH,"ICE4_FAST_RH end:PRRDRYH",PRECISION)
  CALL MPPDB_CHECK1D(PRGDRYH,"ICE4_FAST_RH end:PRGDRYH",PRECISION)
  CALL MPPDB_CHECK1D(PRDRYHG,"ICE4_FAST_RH end:PRDRYHG",PRECISION)
END IF

!$acc end data

!$acc end data

END SUBROUTINE ICE4_FAST_RH
